# This program contains all files relevant to Frequentist Restoration of the Danztzig Path 
#	for the Buckley-James estimator
#
# Brent Johnson
# Emory University
# Department of Biostatistics and Bioinformatics
#
# FUNCTIONS: 
#
# frdspath() : compute the Dantzig path for censored outcomes starting with a 
#	user-supplied initial coefficient estimate.
#
# bjfrdspath() : similar to frdspath but with a preliminary Buckley-James step
#	to determine the active set, then fit full model to active set and use this 
#	coefficient estimate to intialize frdspath.
#
# bjfrds.first() : a routine which implements a variety of different techniques to 
#	produce an initial set of regression coefficients in high- and low-dimensions.
#
# bjest() : Buckley-James estimator for n>p.
#
# gehanest() : Gehan estimator for n>p.
#
# impute() : Buckley-James imputation. NOTE: this function calls a C program that I 
#	source separately in a command window (PC) or terminal application (Mac).  Type:
#	> R CMD SHLIB impute.c
#
#
# USAGE for bjfrdspath:
#
#	INPUT:
#	x = n-by-p matrix of predictors
#	y = n-by-1 vector of f/up time
#	delta = n-by-1 failure indicator
#	actset0 = index to subset entire x matrix.  Using R subscripting so can be either 
#		vector of logicals or vector of integers
#	gehan.flag = use Gehan or BJ estimate as initial estimate when n>p
#	L = number of joints in the path
#	lamvec = alternatively, provide a vector of regularization parms.
#	tol = tolerance for BJ estimation
#	maxit = max iterations
#	standardize = logical to standardize x matrix
#
#	OUTPUT:
#	bpath = L-by-p matrix of coefficients
#	s = l1-norm of coefficient path
#	lam = lambda vector
#
# NOTES: (other notes are below)
#
# 4 Oct 2010 : rewrite to catch error if card(actset0)<1
#
#
# 
bjfrdspath <- function(x,y,delta,actset0,
	gehan.flag=TRUE,
	L=100,
	lamvec=NULL,
	tol=1e-4,
	maxit=10,
	standardize=TRUE,standardize.y=TRUE) {
	
	require(lpSolve)
	require(survival)
	
	x = as.matrix(x)
	n = nrow(x)
	p = ncol(x)
	
	if(alpha<0) {
		stop("alpha must be >= 0")
		}
		
  	if(standardize) {
		my.means <- apply(x,2,mean)
		my.sd <- apply(x,2,sd)
		x <- t((t(x)-my.means)/my.sd)
		}

	# initial estimator on actset0
	# ----------------------------
	ft0 <- bjfrds.first(x,y,delta,actset0,gehan.flag,tol=tol,maxit=maxit,standardize=FALSE)
	
	# frequentist restoration using binit
	# -----------------------------------
	ft <- frdspath(x,y,delta,b0=ft0$coef,L=L,lamvec=lamvec,standardize=FALSE,standardize.y=standardize.y)
	bpath <- ft$path
	lamvec <- ft$lam
	list(path=bpath,s=apply(bpath,1,function(x)sum(abs(x))),lam=lamvec)
	}

frdspath <- function(x,y,delta,b0,L=100,lamvec=NULL,standardize=TRUE,standardize.y=TRUE) {
	
	x <- as.matrix(x)
	n = nrow(x)
	p = ncol(x)
	scale = 1
	#scale = sqrt(2*log(p))

	#if(is.element("lpwrapper",objects())) {
	#	cat("ok\n")
	#	}
	#else {
	#	stop("cannot locate function lpwrapper, source dspath.R")
	#	}

	if(standardize) {
		xavg <- apply(x,2,mean)
		xsd <- apply(x,2,sd)
		x <- t((t(x) - xavg)/xsd)
		}
	
	ynew <- impute(b0,y,delta,x)
	yconst <- 0
	if(standardize.y) {
		yconst <- mean(ynew)
		ynew <- ynew - yconst
		}
	
    colscale <- apply(x,2,function(x){sqrt(sum(x^2))})
    Xstar <- t(t(x)/colscale)
	
    if(is.null(lamvec)) {
		lammax = max(abs(t(Xstar) %*% ynew))
		lammin = 0.001 * lammax
		if (n>p) {
			lamvec = exp(seq(from=log(lammin),to=log(lammax),length=L))
			lamvec[1] = 0
			}
		else {
			lamvec = exp(seq(from=log(lammin),to=log(lammax),length=L))
			}
    	
    	}
    else L <- length(lamvec)
	
	lamorig = lamvec
	lamvec=sort(lamvec,decreasing=TRUE)
	bpath = matrix(unlist(lapply(1:L,lpwrapper,lamvec,ynew,Xstar,scale,colscale)),L,p,byrow=TRUE)
	bpath = bpath[L:1,]
	
	list(path=bpath,s=apply(bpath,1,function(x)sum(abs(x))),lam=lamorig,yconst=yconst)
	}

#
# 28 Sept 2010 : rewrite bjfrds.first to use glmboost as initial active set
# 3 Oct 2010 : rewrite so that no default initial estimator is taken
#
bjfrds.first <- function(x,y,delta,actset0=NULL,gehan.flag=NULL,tol=5e-3,maxit=50,standardize=TRUE) {
	#require(mboost)
	require(survival)
	x <- as.matrix(x)
	n = nrow(x)
	p = ncol(x)
	
  	if(standardize) {
		my.means <- apply(x,2,mean)
		my.sd <- apply(x,2,sd)
		x <- t((t(x)-my.means)/my.sd)
		}

	if(is.null(actset0)) {
		# cross-validate parametric model with extreme value errors
		#ctrl <- boost_control(mstop=500)
		#tmp <- cv.phboost(x,y=exp(y),delta)
		#bph <- tmp$b
		#actset0 <- seq(1:p)[abs(bph)>1E-8]
		stop("no active set supplied with no default procedure")
		}

	p.act = length(actset0)
	
	if(p.act>n) stop("bjfrds.first: active set cannot be larger than n")
	if(p.act>sum(delta)) warning("bjfrds.first: card(actset0) > no. of uncensored obs")
	
	if(p.act<1) { # no active vars
		res=y 
		ft=survfit(Surv(res,delta) ~ 1,se.fit=F)
		S.hat=ft$surv
		UniqRes=ft$time
		bout=rep(0,p)
		}
	else { # p.act >= 1
		if(sum(!(actset0 %in% seq(1,p,by=1)))>0) stop("bjfrds.first: actset0 contains indices outside 1 to p")
		else {
			if((!is.null(gehan.flag))&&(gehan.flag==TRUE)) {
				b0 = gehanest(x[,actset0],y,delta,CENTER=FALSE,exact=TRUE)
				}
			else {
				if((n<100)&&(p.act<20)) {
					# use Gehan estimate if possible
					b0 = gehanest(x[,actset0],y,delta,CENTER=FALSE,exact=TRUE)
					}
				else {
					b0 = lsfit(x[,actset0],y-mean(y),intercept=FALSE)$coef
					b0 = bjest(x[,actset0],y,delta,CENTER=FALSE,tol=tol,maxit=maxit,b0=b0)$coef
					}
				}
			res = y - as.matrix(x[,actset0]) %*% b0
			ft <- survfit(Surv(res,delta) ~ 1,se.fit=F)
			S.hat = ft$surv
			UniqRes = ft$time
			bout=rep(0,p)
			bout[actset0] = b0

			}
		}	

	list(coef=bout,surv=S.hat,time=UniqRes)
	}



bjest <- function(x,y,delta,CENTER=TRUE,tol=1e-4,maxit=10,b0=rep(0,ncol(x))) {
  	#cat("BJ Estimator\n")
  	x <- as.matrix(x)
  	dimnum <- dim(x) 
  	n1 <- dimnum[1]
  	n2 <- dimnum[2]
  	bnew = b0
  	it = 0
  	diff = 2
  	
  	if(CENTER) {
		my.means <- apply(x,2,mean)
		my.sd <- apply(x,2,sd)
		x <- t((t(x)-my.means)/my.sd)
		}
	while((it<maxit)&&(diff>tol)) {
		it <- it + 1
		bold  = bnew
		ynew = impute(bold,y,delta,x)
		ynew = ynew - mean(ynew)
		bnew = lsfit(x,ynew,intercept=FALSE,tol=1e-7)$coef
		diff = max(abs(bnew-bold))
		}
	#cat("end of BJE\n")
	cat("diff = ",round(diff,3),"\n")
	if(any(is.na(bnew)||!is.finite(bnew))) stop("BJ Estimator converged with NAs")
	list(coef=bnew,ytil=(ynew+mean(ynew)))
	}
	
gehanest <- function(x,y,delta,CENTER=TRUE,exact=FALSE,b0=rep(0,ncol(x))) {
	require(quantreg)
	#cat("Gehan Estimator\n")
  	x <- as.matrix(x)
  	dimnum <- dim(x) 
  	n1 <- dimnum[1]
  	n2 <- dimnum[2]
  	if (n1 < n2) {
  		cat("n = ",n1," and p = ",n2,"\n")
  		stop("gehanest: p is greater than n")
  		}
  	if(sum(delta) < n2) warning("gehanest: p is greater than sum(delta)")
  	
  	M=1000*(n1)^2
  	if(CENTER) {
		my.means <- apply(x,2,mean)
		my.sd <- apply(x,2,sd)
		x <- t((t(x)-my.means)/my.sd)
		}
	
  	yy0 <- rep(y,rep(n1,n1))
 	yy1 <- rep(y,n1)
  	yy <- yy0-yy1
  	delta1 <- rep(delta,rep(n1,n1))
  	xx0 <- matrix(rep(as.vector(x),rep(n1,n1*n2)),nrow=n1*n1)
  	xx1 <- t(matrix(rep(as.vector(t(x)),n1),nrow=n2))
  	xx <- xx0-xx1   

  	xxdif <- xx*delta1 
  	xM <- apply(xxdif,2,sum)
  	xnew <- rbind(xxdif,-xM)
  	yynew <- c(delta1*yy,M)

	loc.spline <- function(x,EPS) {
		ifelse(x < -EPS,0,ifelse(x<EPS,-(x+EPS)^3/(4*EPS^3) + (3*(x+EPS)^2)/(4*EPS^2),1))
		}
	loc.wtfun <- function(res,delta1,geps) {
		delta1*loc.spline(-res,geps)/ifelse(abs(res)>0.05,abs(res),0.05)
		}

	loc.irwls <- function(xx,yy,delta1,GEPS,binit=rep(0,ncol(xx))) {
		iter <- 0
		absdiff <- 2
		bnew <- binit
		MAXITER <- 10
		TOL <- 1E-4
		while ((absdiff>TOL)&&(iter<MAXITER)) {
			iter <- iter + 1
			bold <- bnew
			#cat(iter,": b = ",round(bold,3),"\n")
			res <- yy - xx %*% bold
			ww <- loc.wtfun(res,delta1,GEPS)
			xwx <- t(xx) %*% diag(as.vector(ww)) %*% xx
			xwy <- t(xx) %*% diag(as.vector(ww)) %*% yy
			bnew <- qr.solve(xwx,xwy,tol=1e-7)
			absdiff <- max(abs(bnew-bold))
			}
		bnew
		}

  	if(exact) {
  		fit <- rq.fit.fnb(xnew,yynew,tau=0.5)
  		bhat <- fit$coef
  		}
  	else {
  		bhat <- loc.irwls(xx,yy,delta1,GEPS=1e-2,binit=b0)
  		}

  	bhat
}

lpwrapper <- function(J,lamvec,Y,X,lambdap,colscale) {
	out <- linearoptim(Y,X,lamvec[J],lambdap)
	as.vector(out) / colscale
	}

#
# this function from James and Radchencko "Generalized Dantzig estimation .. "
#
"linearoptim" <-
function(Y,X,sigma,lambdap){
    p <- ncol(X)
    f.obj <- rep(1,2*p)
    XX <- t(X)%*%X
    b1 <- lambdap*sigma-as.vector(t(X)%*%Y)
    b2 <- lambdap*sigma+as.vector(t(X)%*%Y)
    f.con <- rbind(-XX,XX)
    f.con <- cbind(f.con,-f.con)
    f.rhs <- c(b1,b2)
    f.dir <- rep("<=",2*p)
    lpout <- lp("min",f.obj,f.con,f.dir,f.rhs)
    gamma <- lpout$solution[1:p]-lpout$solution[-(1:p)]
    if (lpout$status==2) {
    	gamma <- rep(1e4,p)
    	print("Warning: No Solution Found")
    	}
    gamma
    }

impute <- function(b,y,del,x)
{
	require(survival)
	if(!is.loaded("impute")) {
		cat("trying to load impute() object file: impute.so\n")
		#dyn.load("/Users/brentjohnson/Documents/Papers/VarSelection/BJDS/Coding/RCodePublic/impute.so") # full path
		dyn.load("/impute.so") # short path
	}
	
	x <- as.matrix(x)
	n <- length(y)
	mu <- as.vector(x %*% b)
	ei <- y - mu
	
	#ystar <- as.numeric(n)
	ftei <- survfit(Surv(ei,del) ~ 1,se.fit=F)
	Surv <- ftei$surv
	UniqRes <- ftei$time
	nuniq <- length(UniqRes)
	
	out <- .C("impute",
		ystar=double(n),
		as.integer(n),
		as.double(ei),
		as.double(del),
		as.double(mu),
		as.double(Surv),
		as.double(UniqRes),
		as.integer(nuniq))
	#yhat <- out[[1]]
	yhat <- out[["ystar"]]
	invisible(yhat)
}
